/*
 * ngtcp2
 *
 * Copyright (c) 2026 ngtcp2 contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef NGTCP2_WF_H
#define NGTCP2_WF_H

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif /* defined(HAVE_CONFIG_H) */

#include <ngtcp2/ngtcp2.h>

/*
 * ngtcp2_wf implements Kathleen Nichols's windowed min/max tracking
 * algorithm.
 */

typedef struct ngtcp2_wf_sample {
  uint64_t value;
  uint64_t ts;
} ngtcp2_wf_sample;

typedef struct ngtcp2_wf {
  uint64_t win;
  ngtcp2_wf_sample samples[3];
} ngtcp2_wf;

void ngtcp2_wf_init(ngtcp2_wf *wf, uint64_t win);

void ngtcp2_wf_update(ngtcp2_wf *wf, uint64_t value, uint64_t ts);

uint64_t ngtcp2_wf_get_best(const ngtcp2_wf *wf);

#endif /* !defined(NGTCP2_WF_H) */
